import os

from clarifai import __version__

CLIENT_VERSION = __version__
OS_VER = os.sys.platform
PYTHON_VERSION = '.'.join(
    map(str, [os.sys.version_info.major, os.sys.version_info.minor, os.sys.version_info.micro])
)
